import numpy as np
import pytest

from color_correction.utils.yolo_utils import (
    compute_iou,
    multiclass_nms,
    nms,
    xywh2xyxy,
)


# ------------------- Test for nms ---------------------------
@pytest.mark.parametrize(
    "boxes,scores,iou_threshold,expected_len",
    [
        # 1. Happy path - 2 overlapping boxes and 1 separate box
        (
            np.array(
                [[0, 0, 10, 10], [9, 9, 30, 30], [1, 1, 11, 11]],
                dtype=np.float32,
            ),  # noqa: E501
            np.array([0.9, 0.8, 0.7]),
            0.5,
            2,
        ),
        # 2. Edge case - single box
        (
            np.array([[0, 0, 10, 10]], dtype=np.float32),
            np.array([0.9]),
            0.5,
            1,
        ),
        # 3. Edge case - no overlapping boxes
        (
            np.array(
                [[0, 0, 10, 10], [20, 20, 30, 30], [40, 40, 50, 50]],
                dtype=np.float32,
            ),  # noqa: E501
            np.array([0.9, 0.8, 0.7]),
            0.5,
            3,
        ),
        # 4. Edge case - all boxes overlap significantly
        (
            np.array(
                [[0, 0, 10, 10], [1, 1, 11, 11], [2, 2, 10, 10]],
                dtype=np.float32,
            ),
            np.array([0.9, 0.8, 0.7]),
            0.5,
            1,
        ),
        # 5. Edge case - very low IoU threshold
        (
            np.array([[0, 0, 10, 10], [2, 2, 12, 12]], dtype=np.float32),
            np.array([0.9, 0.8]),
            0.1,
            1,
        ),
        # 6. Edge case - very high IoU threshold
        (
            np.array([[0, 0, 10, 10], [2, 2, 12, 12]], dtype=np.float32),
            np.array([0.9, 0.8]),
            0.9,
            2,
        ),
    ],
)
def test_nms_parametrized(
    boxes: np.ndarray,
    scores: np.ndarray,
    iou_threshold: float,
    expected_len: int,
) -> None:
    keep_indices = nms(boxes, scores, iou_threshold)
    assert len(keep_indices) == expected_len
    assert 0 in keep_indices  # Highest scoring box should always be kept
    assert all(
        scores[i] >= scores[j]
        for i, j in zip(keep_indices[:-1], keep_indices[1:], strict=False)
    )  # Scores should be descending


def test_nms_empty() -> None:
    empty_boxes = np.array([], dtype=np.float32).reshape(0, 4)
    empty_scores = np.array([], dtype=np.float32)
    keep_indices = nms(boxes=empty_boxes, scores=empty_scores, iou_threshold=0.5)
    assert len(keep_indices) == 0


def test_nms_same_score() -> None:
    boxes = np.array([[0, 0, 10, 10], [20, 20, 30, 30]], dtype=np.float32)
    scores = np.array([0.9, 0.9])
    keep_indices = nms(boxes=boxes, scores=scores, iou_threshold=0.5)
    assert len(keep_indices) == 2


# ------------------- Test for compute_iou -------------------
@pytest.mark.parametrize(
    "box,boxes,expected_ious",
    [
        # No overlap case
        (
            np.array([0, 0, 10, 10]),
            np.array([[20, 20, 30, 30], [40, 40, 50, 50]]),
            np.array([0.0, 0.0]),
        ),
        # Partial overlap case
        (
            np.array([0, 0, 20, 20]),
            np.array([[10, 10, 30, 30]]),
            np.array([0.14285714]),
        ),
        # Complete overlap (identical boxes)
        (
            np.array([0, 0, 10, 10]),
            np.array([[0, 0, 10, 10]]),
            np.array([1.0]),
        ),
        # Multiple boxes with varying overlap
        (
            np.array([0, 0, 10, 10]),
            np.array([[5, 5, 15, 15], [0, 0, 5, 5], [20, 20, 30, 30]]),
            np.array([0.14285714, 0.25, 0.0]),
        ),
    ],
)
def test_compute_iou(
    box: np.ndarray,
    boxes: np.ndarray,
    expected_ious: np.ndarray,
) -> None:
    ious = compute_iou(box, boxes)
    np.testing.assert_array_almost_equal(ious, expected_ious, decimal=7)


def test_compute_iou_empty_boxes() -> None:
    box = np.array([0, 0, 10, 10])
    empty_boxes = np.array([], dtype=np.float32).reshape(0, 4)
    ious = compute_iou(box, empty_boxes)
    assert len(ious) == 0


# ------------------- Test for multiclass_nms -------------------
@pytest.mark.parametrize(
    "boxes,scores,class_ids,iou_threshold,expected_len",
    [
        # 1. Happy path - multiple classes with some overlapping boxes
        (
            np.array(
                [[0, 0, 10, 10], [1, 1, 11, 11], [20, 20, 30, 30]],
                dtype=np.float32,
            ),
            np.array([0.9, 0.8, 0.7]),
            np.array([0, 0, 1]),
            0.5,
            2,
        ),
        # 2. Edge case - single class
        (
            np.array([[0, 0, 10, 10], [1, 1, 11, 11]], dtype=np.float32),
            np.array([0.9, 0.8]),
            np.array([0, 0]),
            0.5,
            1,
        ),
        # 3. Edge case - no overlapping boxes, different classes
        (
            np.array(
                [[0, 0, 10, 10], [20, 20, 30, 30], [40, 40, 50, 50]],
                dtype=np.float32,
            ),
            np.array([0.9, 0.8, 0.7]),
            np.array([0, 1, 2]),
            0.5,
            3,
        ),
        # 4. Edge case - multiple overlapping boxes with same class
        (
            np.array(
                [[0, 0, 10, 10], [1, 1, 11, 11], [2, 2, 10, 10]],
                dtype=np.float32,
            ),
            np.array([0.9, 0.8, 0.7]),
            np.array([0, 0, 0]),
            0.5,
            1,
        ),
        # 5. Edge case - multiple overlapping boxes with different classes
        (
            np.array(
                [[0, 0, 10, 10], [1, 1, 11, 11], [2, 2, 12, 12]],
                dtype=np.float32,
            ),
            np.array([0.9, 0.8, 0.7]),
            np.array([0, 1, 2]),
            0.5,
            3,
        ),
    ],
)
def test_multiclass_nms_parametrized(
    boxes: np.ndarray,
    scores: np.ndarray,
    class_ids: np.ndarray,
    iou_threshold: float,
    expected_len: int,
) -> None:
    keep_indices = multiclass_nms(boxes, scores, class_ids, iou_threshold)
    assert len(keep_indices) == expected_len

    # Check that boxes with highest scores per class are kept
    unique_classes = np.unique(class_ids)
    for class_id in unique_classes:
        class_mask = class_ids == class_id
        class_scores = scores[class_mask]
        if len(class_scores) > 0:
            max_score_idx = np.where(class_mask)[0][np.argmax(class_scores)]
            assert max_score_idx in keep_indices


def test_multiclass_nms_empty() -> None:
    empty_boxes = np.array([], dtype=np.float32).reshape(0, 4)
    empty_scores = np.array([], dtype=np.float32)
    empty_class_ids = np.array([], dtype=np.int32)
    keep_indices = multiclass_nms(
        boxes=empty_boxes,
        scores=empty_scores,
        class_ids=empty_class_ids,
        iou_threshold=0.5,
    )
    assert len(keep_indices) == 0


def test_multiclass_nms_same_score_diff_class() -> None:
    boxes = np.array([[0, 0, 10, 10], [20, 20, 30, 30]], dtype=np.float32)
    scores = np.array([0.9, 0.9])
    class_ids = np.array([0, 1])
    keep_indices = multiclass_nms(
        boxes=boxes,
        scores=scores,
        class_ids=class_ids,
        iou_threshold=0.5,
    )
    assert len(keep_indices) == 2


# ------------------- Test for xywh2xyxy -------------------
@pytest.mark.parametrize(
    "input_boxes,expected_boxes",
    [
        # Single box test case
        (
            np.array([[10, 10, 20, 20]], dtype=np.float32),
            np.array([[0, 0, 20, 20]], dtype=np.float32),
        ),
        # Multiple boxes test case
        (
            np.array([[10, 10, 20, 20], [30, 30, 40, 40]], dtype=np.float32),
            np.array([[0, 0, 20, 20], [10, 10, 50, 50]], dtype=np.float32),
        ),
        # Test with decimal values
        (
            np.array([[5.5, 5.5, 3, 3]], dtype=np.float32),
            np.array([[4, 4, 7, 7]], dtype=np.float32),
        ),
        # Test with zero width/height
        (
            np.array([[10, 10, 0, 0]], dtype=np.float32),
            np.array([[10, 10, 10, 10]], dtype=np.float32),
        ),
    ],
)
def test_xywh2xyxy_parametrized(
    input_boxes: np.ndarray,
    expected_boxes: np.ndarray,
) -> None:
    output_boxes = xywh2xyxy(input_boxes)
    np.testing.assert_array_almost_equal(output_boxes, expected_boxes)


def test_xywh2xyxy_empty() -> None:
    empty_boxes = np.array([], dtype=np.float32).reshape(0, 4)
    output = xywh2xyxy(empty_boxes)
    assert output.shape == (0, 4)



import time

import numpy as np
from sklearn.linear_model import LinearRegression

from color_correction.core.correction.base import BaseComputeCorrection
from color_correction.utils.correction import (
    postprocessing_compute,
    preprocessing_compute,
)


class LinearReg(BaseComputeCorrection):
    def __init__(self) -> None:
        self.model = None

    def fit(
        self,
        x_patches: np.ndarray,  # input patches
        y_patches: np.ndarray,  # reference patches
    ) -> np.ndarray:
        start_time = time.perf_counter()

        self.model = LinearRegression(fit_intercept=False).fit(x_patches, y_patches)

        exc_time = time.perf_counter() - start_time
        print(f"{self.__class__.__name__} Fit: {exc_time} seconds")
        return self.model

    def compute_correction(self, input_image: np.ndarray) -> np.ndarray:
        if self.model is None:
            raise ValueError("Model is not fitted yet. Please call fit() method first.")

        org_input_shape = input_image.shape
        input_image = preprocessing_compute(input_image)
        image = self.model.predict(input_image)
        corrected_image = postprocessing_compute(org_input_shape, image)
        return corrected_image



import json
import numpy as np
import pytest
from color_correction.utils.formater import format_value

@pytest.mark.parametrize("value,expected_check", [
    # For numpy array image: expecting an <img src="..."/> tag.
    (np.array([[1, 2, 3]], dtype=np.uint8), lambda out: out.startswith('<img src="')),
    # For dict: expecting a valid json string that decodes back to the same dict.
    ({"a": 1, "b": 2}, lambda out: json.loads(out) == {"a": 1, "b": 2}),
    # For list: expecting a valid json string that decodes back to the same list.
    ([1, 2, 3], lambda out: json.loads(out) == [1, 2, 3]),
    # For numpy float types: expecting formatted float with 4 decimals.
    (np.float64(3.14159265), lambda out: out == "3.1416"),
    (np.float32(2.71828), lambda out: out == "2.7183"),
    # For built-in float: the function doesn't format, returns str(value)
    (2.0, lambda out: out == "2.0"),
    # For plain string: should return same string.
    ("test string", lambda out: out == "test string"),
])
def test_format_value(value, expected_check):
    result = format_value(value)
    assert expected_check(result)



import time

import cv2
import numpy as np
import onnxruntime

from color_correction.core.card_detection.base import BaseCardDetector
from color_correction.schemas.det_yv8 import DetectionResult
from color_correction.utils.downloader import downloader_model_yolov8
from color_correction.utils.yolo_utils import (
    multiclass_nms,
    xywh2xyxy,
)


class YOLOv8CardDetector(BaseCardDetector):
    """YOLOv8CardDetector is a class that implements card detection
    using the YOLOv8 model.

    Reference
    ---------
    https://github.com/ibaiGorordo/ONNX-YOLOv8-Object-Detection/blob/main/yolov8/YOLOv8.py

    """

    def __init__(
        self,
        conf_th: float = 0.15,
        iou_th: float = 0.7,
        path: str | None = None,
        use_gpu: bool = False,
    ) -> None:
        self.conf_threshold = conf_th
        self.iou_threshold = iou_th
        self.use_gpu = use_gpu
        if path is None:
            path = downloader_model_yolov8(use_gpu)
        self.__initialize_model(path)

    def detect(self, image: np.ndarray) -> DetectionResult:
        """
        Detect objects in the given image using YOLOv8 model.

        Parameters
        ----------
        image : np.ndarray
            The input image BGR in which to detect objects.

        Returns
        -------
        DetectionResult
            A dataclass containing detected bounding boxes, confidence scores,
            and class IDs.
        """
        input_tensor = self.__prepare_input(image.copy())
        outputs = self.__inference(input_tensor)
        boxes, scores, class_ids = self.__process_output(outputs)

        det_res = DetectionResult(
            boxes=boxes,
            scores=scores,
            class_ids=class_ids,
        )

        return det_res

    # Service functions
    def __initialize_model(self, path: str) -> None:
        self.session = onnxruntime.InferenceSession(
            path,
            providers=onnxruntime.get_available_providers(),
        )
        # Get model info
        self.__get_input_details()
        self.__get_output_details()

    def __prepare_input(self, original_image: np.ndarray) -> np.ndarray:
        [height, width, _] = original_image.shape

        # expected shape based on model input
        expected_width = self.input_width
        expected_height = self.input_height
        expected_length = min((expected_height, expected_width))

        length = max((height, width))
        # self.scale_to_expected = expected_length / length
        self.scale_to_ori = length / expected_length

        image = np.zeros((length, length, 3), np.uint8)
        image[0:height, 0:width] = original_image

        input_image = cv2.resize(image, (expected_width, expected_height))

        if self.use_gpu:
            input_image = (input_image / 255.0).astype(np.float16)
        else:
            input_image = (input_image / 255.0).astype(np.float32)
        # Channel first
        input_image = input_image.transpose(2, 0, 1)

        # Expand dimensions
        input_image = np.expand_dims(input_image, axis=0)
        return input_image

    def __inference(self, input_tensor: np.ndarray) -> list[np.ndarray]:
        start = time.perf_counter()  # noqa: F841
        outputs = self.session.run(
            self.output_names,
            {self.input_names[0]: input_tensor},
        )

        # print(f"Inference time: {(time.perf_counter() - start) * 1000:.2f} ms")
        return outputs

    def __process_output(
        self,
        output: list[np.ndarray],
    ) -> tuple[list[list[int]], list[float], list[int]]:
        predictions = np.squeeze(output[0]).T

        # Filter out object confidence scores below threshold
        scores = np.max(predictions[:, 4:], axis=1)
        predictions = predictions[scores > self.conf_threshold, :]
        scores = scores[scores > self.conf_threshold]

        if len(scores) == 0:
            return [], [], []

        # Get the class with the highest confidence
        class_ids = np.argmax(predictions[:, 4:], axis=1)

        # Get bounding boxes for each object
        boxes = self.__extract_boxes(predictions)

        # Apply non-maxima suppression to suppress weak, overlapping bounding boxes
        # indices = nms(boxes, scores, self.iou_threshold)
        indices = multiclass_nms(
            boxes,
            scores,
            class_ids,
            self.iou_threshold,
        )

        return (
            boxes[indices].astype(int).tolist(),
            scores[indices].tolist(),
            class_ids[indices].tolist(),
        )

    # Helper functions
    def __extract_boxes(self, predictions: np.ndarray) -> np.ndarray:
        # Extract boxes from predictions
        boxes = predictions[:, :4]

        # Scale boxes to original image dimensions
        boxes = self.__rescale_boxes(boxes)

        # Convert boxes to xyxy format
        boxes = xywh2xyxy(boxes)

        return boxes

    def __rescale_boxes(self, boxes: np.ndarray) -> np.ndarray:
        # Rescale boxes to original image dimensions
        boxes *= self.scale_to_ori
        return boxes

    def __get_input_details(self) -> None:
        model_inputs = self.session.get_inputs()
        self.input_names = [model_inputs[i].name for i in range(len(model_inputs))]

        self.input_shape = model_inputs[0].shape
        self.input_height = self.input_shape[2]
        self.input_width = self.input_shape[3]

    def __get_output_details(self) -> None:
        model_outputs = self.session.get_outputs()
        self.output_names = [model_outputs[i].name for i in range(len(model_outputs))]


if __name__ == "__main__":
    print("YOLOv8CardDetector")
    model_path = "color_correction/asset/.model/yv8-det.onnx"
    image_path = "color_correction/asset/images/cc-1.jpg"
    image_path = "color_correction/asset/images/Test 19.png"
    detector = YOLOv8CardDetector(conf_th=0.15, iou_th=0.7, use_gpu=True)

    input_image = cv2.imread(image_path)
    # input_image = cv2.resize(input_image, (640, 640))
    result = detector.detect(input_image)
    result.print_summary()
    image_drawed = result.draw_detections(input_image)
    cv2.imwrite("result.png", image_drawed)



__version__ = "0.0.1b3"

# fmt: off
from .constant.color_checker import reference_color_d50_bgr as REFERENCE_COLOR_D50_BGR  # noqa: N812, I001
from .constant.color_checker import reference_color_d50_rgb as REFERENCE_COLOR_D50_RGB  # noqa: N812, I001
from .core.card_detection.det_yv8_onnx import YOLOv8CardDetector
from .schemas.det_yv8 import DetectionResult as YOLOv8DetectionResult
from .services.color_correction import ColorCorrection
from .services.correction_analyzer import ColorCorrectionAnalyzer
# fmt: on

__all__ = [
    "__version__",
    "REFERENCE_COLOR_D50_BGR",
    "REFERENCE_COLOR_D50_RGB",
    "ColorCorrection",
    "ColorCorrectionAnalyzer",
    "YOLOv8CardDetector",
    "YOLOv8DetectionResult",
]



import numpy as np
import shapely

box_tuple = tuple[int, int, int, int]


def get_max_iou_shapely(
    ref_box: shapely.geometry.box,
    target_boxes: list[shapely.geometry.box],
) -> tuple[float, int, shapely.geometry.box]:
    max_iou = 0
    max_idx = -1

    # Compare with each target box
    for idx, target_box in enumerate(target_boxes):
        # Calculate intersection and union
        intersection_area = ref_box.intersection(target_box).area
        union_area = ref_box.union(target_box).area

        # Calculate IoU
        iou = intersection_area / union_area if union_area > 0 else 0

        # Update maximum IoU if current is larger
        if iou > max_iou:
            max_iou = iou
            max_idx = idx

    return max_iou, max_idx, target_boxes[max_idx]


def box_to_xyxy(box: shapely.geometry.box) -> tuple[int, int, int, int]:
    """Convert shapely box to xyxy format"""
    minx, miny, maxx, maxy = box.bounds
    return int(minx), int(miny), int(maxx), int(maxy)


def box_centroid_xy(box: shapely.geometry.box) -> tuple[int, int]:
    return int(box.centroid.x), int(box.centroid.y)


def generate_expected_patches(card_box: box_tuple) -> list[box_tuple]:
    card_x1, card_y1, card_x2, card_y2 = card_box
    card_width = card_x2 - card_x1
    card_height = card_y2 - card_y1

    # get expected grid of cards
    patch_width = card_width / 6
    patch_height = card_height / 4

    expected_patches = []
    for row in range(4):
        for col in range(6):
            x1 = int(card_x1 + col * patch_width)
            y1 = int(card_y1 + row * patch_height)
            x2 = int(x1 + patch_width)
            y2 = int(y1 + patch_height)
            expected_patches.append((x1, y1, x2, y2))

    return expected_patches


def extract_intersecting_patches(
    ls_patches: list[box_tuple],
    ls_grid_card: list[box_tuple],
) -> list[tuple[box_tuple, tuple[int, int]]]:
    ls_ordered_patch = []
    for _, grid_card in enumerate(ls_grid_card, start=1):
        # get intesect patch
        gx1, gy1, gx2, gy2 = grid_card
        grid_box = shapely.box(*grid_card)
        ls_intersect = [
            shapely.box(*xyxy)
            for xyxy in ls_patches
            if grid_box.intersects(shapely.box(*xyxy))
        ]
        len_intersect = len(ls_intersect)
        if len_intersect > 0:
            max_iou, max_id, intersect_box = get_max_iou_shapely(
                ref_box=grid_box,
                target_boxes=ls_intersect,
            )
            # intersect_box = ls_intersect[max_id]
            val = box_to_xyxy(intersect_box)
            xy = box_centroid_xy(intersect_box)
            ls_ordered_patch.append((val, xy))
        else:
            ls_ordered_patch.append(None)
    return ls_ordered_patch


def calculate_patch_statistics(ls_ordered_patch: list[box_tuple]) -> tuple:
    ls_dx = []
    ls_dy = []
    ls_w_grid = []
    ls_h_grid = []
    for idx, patch in enumerate(ls_ordered_patch):
        if patch is None:
            continue

        ls_w_grid.append(patch[2] - patch[0])
        ls_h_grid.append(patch[3] - patch[1])

        if idx not in [5, 11, 17, 23] or idx == 0:
            x1 = patch[0]
            next_x1 = ls_ordered_patch[idx + 1]
            if next_x1 is not None:
                dx = next_x1[0] - x1
                ls_dx.append(dx)

        syarat = idx + 6
        if syarat < len(ls_ordered_patch):
            y1 = patch[1]
            next_y1 = ls_ordered_patch[idx + 6]
            if next_y1 is not None:
                dy = next_y1[1] - y1
                ls_dy.append(dy)

    mean_dx = np.mean(ls_dx)
    mean_dy = np.mean(ls_dy)
    mean_w = np.mean(ls_w_grid)
    mean_h = np.mean(ls_h_grid)

    return mean_dx, mean_dy, mean_w, mean_h


def suggest_missing_patch_coordinates(  # noqa: C901
    ls_ordered_patch: list[box_tuple],
) -> dict[int, box_tuple]:
    d_suggest = {}

    mean_dx, mean_dy, mean_w, mean_h = calculate_patch_statistics(
        ls_ordered_patch=ls_ordered_patch,
    )

    for idx, patch in enumerate(ls_ordered_patch):
        if patch is not None:
            continue

        # looking for nearest neghbor
        neigh_right = None
        neigh_left = None
        neigh_top = None
        neigh_bottom = None

        id_neigh_right = idx + 1
        id_neigh_left = idx - 1
        id_neigh_top = idx - 6
        id_neigh_bottom = idx + 6

        if id_neigh_right not in [0, 6, 12, 18] and id_neigh_right <= 23:
            neigh_right = ls_ordered_patch[id_neigh_right]

        if id_neigh_left not in [5, 11, 17, 23] and id_neigh_left >= 0:
            neigh_left = ls_ordered_patch[id_neigh_left]

        if id_neigh_top >= 0:
            neigh_top = ls_ordered_patch[id_neigh_top]

        if id_neigh_bottom <= 23:
            neigh_bottom = ls_ordered_patch[id_neigh_bottom]

        suggested_patch = None

        # print(f"neigh_right: {neigh_right}")
        # print(f"neigh_left: {neigh_left}")
        # print(f"neigh_top: {neigh_top}")
        # print(f"neigh_bottom: {neigh_bottom}")

        if neigh_right is not None:
            # Dari kanan, geser ke kiri dengan mean_dx
            x1 = neigh_right[0] - mean_dx
            y1 = neigh_right[1]
            suggested_patch = (
                int(x1),
                int(y1),
                int(x1 + mean_w),
                int(y1 + mean_h),
            )

        elif neigh_left is not None:
            # Dari kiri, geser ke kanan dengan mean_dx
            x1 = neigh_left[0] + int(mean_dx)
            y1 = neigh_left[1]
            suggested_patch = (
                int(x1),
                int(y1),
                int(x1 + mean_w),
                int(y1 + mean_h),
            )

        elif neigh_top is not None:
            # Dari atas, geser ke bawah dengan mean_dy
            x1 = neigh_top[0]
            y1 = neigh_top[1] + mean_dy
            suggested_patch = (
                int(x1),
                int(y1),
                int(x1 + mean_w),
                int(y1 + mean_h),
            )

        elif neigh_bottom is not None:
            # Dari bawah, geser ke atas dengan mean_dy
            x1 = neigh_bottom[0]
            y1 = neigh_bottom[1] - mean_dy
            suggested_patch = (
                int(x1),
                int(y1),
                int(x1 + mean_w),
                int(y1 + mean_h),
            )

        d_suggest[idx] = suggested_patch

    return d_suggest



import json

import numpy as np

from color_correction.utils.image_processing import numpy_array_to_base64


def format_value(value: np.ndarray | dict | list | float | str) -> str:
    """Format different types of values for HTML display"""
    if isinstance(value, np.ndarray):  # Image arrays
        return f'<img src="{numpy_array_to_base64(value, convert_bgr_to_rgb=True)}"/>'
    elif isinstance(value, dict | list):  # Dictionaries or lists
        return json.dumps(value)
    elif isinstance(value, np.float64 | np.float32):  # Numpy float types
        return f"{float(value):.4f}"
    return str(value)



"""YOLOv8 utility functions for object detection tasks.

This module provides utility functions for processing YOLOv8 detections, including:
- Non-maximum suppression (NMS)
- Bounding box conversions
- Drawing functions for visualization
"""

import cv2
import numpy as np

from color_correction.constant.yolov8_det import class_names

# Constants
RANDOM_SEED = 3
DEFAULT_COLOR = (0, 0, 255)
DEFAULT_THICKNESS = 2
DEFAULT_MASK_ALPHA = 0.3

# Initialize colors for visualization
rng = np.random.default_rng(RANDOM_SEED)
COLORS = rng.uniform(0, 255, size=(len(class_names), 3))


# Detection Processing Functions
def nms(boxes: np.ndarray, scores: np.ndarray, iou_threshold: float) -> list[int]:
    """Apply Non-Maximum Suppression (NMS) to filter overlapping bounding boxes.

    NMS is used to eliminate redundant bounding boxes in object detection tasks.
    It selects the bounding boxes with the highest confidence scores while
    removing boxes that overlap significantly with them.

    Logic:
    - No Overlap (IoU = 0):
      Boxes are retained as they do not affect each other.
    - High Overlap (IoU > threshold):
      The box with the lower confidence score is removed, as it is considered
      a duplicate detection of the same object.
    - Low Overlap (IoU < threshold):
      Both boxes are retained, as they are considered detections of different objects.

    Parameters
    ----------
    boxes : np.ndarray
        Array of bounding boxes in format (x1, y1, x2, y2).
    scores : np.ndarray
        Array of confidence scores for each box.
    iou_threshold : float
        IoU threshold for filtering overlapping boxes.

    Returns
    -------
    list[int]
        Indices of the bounding boxes to keep after applying NMS.
    """
    # Sort by score
    sorted_indices = np.argsort(scores)[::-1]
    keep_boxes = []

    while sorted_indices.size > 0:
        # Pick the last box
        box_id = sorted_indices[0]
        keep_boxes.append(box_id)

        # if just only one box, then break
        if sorted_indices.size == 1:
            break

        # Compute IoU of the picked box with the rest
        ious = compute_iou(boxes[box_id, :], boxes[sorted_indices[1:], :])

        # Remove boxes with IoU over the threshold,
        keep_indices = np.where(ious < iou_threshold)[0]

        # update sorted_indices
        sorted_indices = sorted_indices[1:][keep_indices]

    return keep_boxes


def multiclass_nms(
    boxes: np.ndarray,
    scores: np.ndarray,
    class_ids: np.ndarray,
    iou_threshold: float,
) -> list[int]:
    """Apply non-maximum suppression to boxes across multiple classes.

    Parameters
    ----------
    boxes : np.ndarray
        Array of bounding boxes in format (x1, y1, x2, y2).
    scores : np.ndarray
        Array of confidence scores for each box.
    class_ids : np.ndarray
        Array of class IDs for each box.
    iou_threshold : float
        IoU threshold for filtering overlapping boxes.

    Returns
    -------
    list[int]
        Indices of boxes to keep after multiclass NMS.
    """
    unique_class_ids = np.unique(class_ids)

    keep_boxes = []
    for class_id in unique_class_ids:
        class_indices = np.where(class_ids == class_id)[0]
        class_boxes = boxes[class_indices, :]
        class_scores = scores[class_indices]

        class_keep_boxes = nms(class_boxes, class_scores, iou_threshold)
        keep_boxes.extend(class_indices[class_keep_boxes])
    return keep_boxes


def compute_iou(box: np.ndarray, boxes: np.ndarray) -> np.ndarray:
    """Compute Intersection over Union between a box and an array of boxes.

    Parameters
    ----------
    box : np.ndarray
        Single box in format (x1, y1, x2, y2).
    boxes : np.ndarray
        Array of boxes to compute IoU with.

    Returns
    -------
    np.ndarray
        Array of IoU values between the input box and each box in boxes.
    """
    box = box.astype(np.float32)
    boxes = boxes.astype(np.float32)

    # Compute xmin, ymin, xmax, ymax for both boxes
    xmin = np.maximum(box[0], boxes[:, 0])
    ymin = np.maximum(box[1], boxes[:, 1])
    xmax = np.minimum(box[2], boxes[:, 2])
    ymax = np.minimum(box[3], boxes[:, 3])

    # Compute intersection area
    intersection_area = np.maximum(0, xmax - xmin) * np.maximum(0, ymax - ymin)

    # Compute union area
    box_area = (box[2] - box[0]) * (box[3] - box[1])
    boxes_area = (boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1])
    union_area = box_area + boxes_area - intersection_area

    # Compute IoU
    iou = intersection_area / union_area
    return iou


def xywh2xyxy(x: np.ndarray) -> np.ndarray:
    """Convert bounding box format from (x, y, w, h) to (x1, y1, x2, y2).

    Parameters
    ----------
    x : np.ndarray
        Array of boxes in (x, y, w, h) format.

    Returns
    -------
    np.ndarray
        Array of boxes in (x1, y1, x2, y2) format.
    """
    # Convert bounding box (x, y, w, h) to bounding box (x1, y1, x2, y2)
    y = np.copy(x)
    y[..., 0] = x[..., 0] - x[..., 2] / 2
    y[..., 1] = x[..., 1] - x[..., 3] / 2
    y[..., 2] = x[..., 0] + x[..., 2] / 2
    y[..., 3] = x[..., 1] + x[..., 3] / 2
    return y


# Visualization Functions
def draw_detections(
    image: np.ndarray,
    boxes: list[list[int]],
    scores: list[float],
    class_ids: list[int],
    mask_alpha: float = DEFAULT_MASK_ALPHA,
) -> np.ndarray:
    """Draw detection boxes, labels and masks on the image.

    Parameters
    ----------
    image : np.ndarray
        Input image to draw on.
    boxes : list[list[int]]
        List of bounding boxes in format (x1, y1, x2, y2).
    scores : list[float]
        List of confidence scores for each detection.
    class_ids : list[int]
        List of class IDs for each detection.
    mask_alpha : float, optional
        Transparency of the mask overlay, by default DEFAULT_MASK_ALPHA.

    Returns
    -------
    np.ndarray
        Image with drawn detections.
    """
    det_img = image.copy()
    img_height, img_width = image.shape[:2]

    # Calculate font and text properties based on image size
    font_size = min([img_height, img_width]) * 0.0005
    text_thickness = int(min([img_height, img_width]) * 0.001)

    # Draw masks first (background layer)
    det_img = draw_masks(det_img, boxes, class_ids, mask_alpha)

    # Draw boxes and labels (foreground layer)
    for class_id, box, score in zip(class_ids, boxes, scores, strict=False):
        color = COLORS[class_id]
        draw_box(det_img, box, color)

        # Create and draw label
        label = f"{class_names[class_id]} {int(score * 100)}%"
        draw_text(det_img, label, box, color, font_size, text_thickness)

    return det_img


def draw_box(
    image: np.ndarray,
    box: list[int],
    color: tuple[int, int, int] = DEFAULT_COLOR,
    thickness: int = DEFAULT_THICKNESS,
) -> np.ndarray:
    """Draw a bounding box on the image.

    Parameters
    ----------
    image : np.ndarray
        Input image to draw on.
    box : list[int]
        Bounding box coordinates in format (x1, y1, x2, y2).
    color : tuple[int, int, int], optional
        RGB color for the box, by default DEFAULT_COLOR.
    thickness : int, optional
        Line thickness of the box, by default DEFAULT_THICKNESS.

    Returns
    -------
    np.ndarray
        Image with drawn box.
    """
    x1, y1, x2, y2 = box
    return cv2.rectangle(image, (x1, y1), (x2, y2), color, thickness)


def draw_text(
    image: np.ndarray,
    text: str,
    box: list[int],
    color: tuple[int, int, int] = DEFAULT_COLOR,
    font_size: float = 0.001,
    text_thickness: int = 2,
) -> np.ndarray:
    """Draw text with background on the image.

    Parameters
    ----------
    image : np.ndarray
        Input image to draw on.
    text : str
        Text to be drawn.
    box : list[int]
        Bounding box coordinates where text will be placed.
    color : tuple[int, int, int], optional
        RGB color for text background, by default DEFAULT_COLOR.
    font_size : float, optional
        Size of the font, by default 0.001.
    text_thickness : int, optional
        Thickness of the text, by default 2.

    Returns
    -------
    np.ndarray
        Image with drawn text.
    """
    x1, y1, x2, y2 = box
    (tw, th), _ = cv2.getTextSize(
        text=text,
        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
        fontScale=font_size,
        thickness=text_thickness,
    )
    th = int(th * 1.2)

    cv2.rectangle(image, (x1, y1), (x1 + tw, y1 - th), color, -1)

    return cv2.putText(
        image,
        text,
        (x1, y1),
        cv2.FONT_HERSHEY_SIMPLEX,
        font_size,
        (255, 255, 255),
        text_thickness,
        cv2.LINE_AA,
    )


def draw_masks(
    image: np.ndarray,
    boxes: list[list[int]],
    classes: list[int],
    mask_alpha: float = DEFAULT_MASK_ALPHA,
) -> np.ndarray:
    """Draw semi-transparent masks for detection boxes.

    Parameters
    ----------
    image : np.ndarray
        Input image to draw on.
    boxes : list[list[int]]
        List of bounding boxes in format (x1, y1, x2, y2).
    classes : list[int]
        List of class IDs for each box.
    mask_alpha : float, optional
        Transparency of the mask overlay, by default DEFAULT_MASK_ALPHA.

    Returns
    -------
    np.ndarray
        Image with drawn masks.
    """
    mask_img = image.copy()

    # Draw bounding boxes and labels of detections
    for box, class_id in zip(boxes, classes, strict=False):
        color = COLORS[class_id]

        x1, y1, x2, y2 = box

        # Draw fill rectangle in mask image
        cv2.rectangle(mask_img, (x1, y1), (x2, y2), color, -1)

    return cv2.addWeighted(mask_img, mask_alpha, image, 1 - mask_alpha, 0)



from abc import ABC, abstractmethod

import numpy as np

from color_correction.schemas.det_yv8 import DetectionResult


class BaseCardDetector(ABC):
    @abstractmethod
    def detect(
        self,
        image: np.ndarray,
        conf: float = 0.15,
    ) -> DetectionResult: ...



from color_correction.core.correction._factory import (
    CorrectionModelFactory,
)
from color_correction.core.correction.affine_reg import AffineReg
from color_correction.core.correction.least_squares import (
    LeastSquaresRegression,
)
from color_correction.core.correction.linear_reg import LinearReg
from color_correction.core.correction.polynomial import Polynomial

__all__ = [
    "CorrectionModelFactory",
    "LeastSquaresRegression",
    "Polynomial",
    "LinearReg",
    "AffineReg",
]



import os
import re
from typing import Final

import httpx

from color_correction.schemas.device import GPUType
from color_correction.utils.device_info import get_device_specs


def download_google_drive_file(file_id: str, output_file: str) -> None:
    """
    Download a file from Google Drive using its file ID.

    Parameters
    ----------
    file_id : str
        The unique ID of the file on Google Drive.
    output_file : str
        The name of the output file where the content will be saved.

    Returns
    -------
    None
        Downloads the file and saves it locally.
    """
    url: Final = f"https://drive.google.com/uc?export=download&id={file_id}"

    with httpx.Client(follow_redirects=True, timeout=60) as client:
        print(f"Start downloading file from: {url}")
        # First request to get confirmation token
        response = client.get(url)
        print(f"Response status code: {response.status_code}")
        if response.status_code == 303 or "download_warning" in response.text:
            # Handle large file confirmation
            print("Large file confirmation needed if file is not downloaded.")
            confirm_token = re.findall(r"confirm=([^&]+)", response.url.query)
            print(f"Confirm token: {confirm_token}")
            if confirm_token:
                url = f"{url}&confirm={confirm_token[0]}"
                print(f"New URL: {url}")

        # Stream download
        with client.stream("GET", url, timeout=60) as response:
            print(f"Downloading file to: {output_file}")
            response.raise_for_status()
            with open(output_file, "wb") as f:
                for chunk in response.iter_bytes():
                    f.write(chunk)
        print(f"File downloaded: {output_file}")


def downloader_model_yolov8(use_gpu: bool = False) -> str:
    specs = get_device_specs()
    model_folder = os.path.join(os.getcwd(), "tmp", "models")
    if use_gpu:
        if specs.is_apple_silicon:
            print("Apple Silicon device detected.")
            fileid = "19O02x_Co2ceBQHHNUib-aVp0cjgG9V-r"
            filename = "yv8-det-mps.onnx"
        elif specs.gpu_type == GPUType.NVIDIA:
            print("NVIDIA GPU detected.")
            fileid = "19O02x_Co2ceBQHHNUib-aVp0cjgG9V-r"
            filename = "yv8-det-gpu.onnx"
        else:
            raise ValueError(
                "GPU not detected or not supported. "
                "Please use CPU device. Device Info: ",
                specs,
            )
    else:
        print("CPU device detected.")
        fileid = "1d1p2HCltiJeVGi6NtDanJvqLbzXs8NLP"
        filename = "yv8-det-cpu.onnx"

    os.makedirs(model_folder, exist_ok=True)
    fullpath = os.path.join(model_folder, filename)
    if os.path.exists(fullpath):
        return fullpath
    print("Auto downloading YOLOv8 model...")
    download_google_drive_file(fileid, fullpath)
    return fullpath


if __name__ == "__main__":
    # Example usage
    # file_id = "19O02x_Co2ceBQHHNUib-aVp0cjgG9V-r"
    # output_file = "yv8-det-mps.onnx"
    filename = downloader_model_yolov8()
    print(f"Output file name: {filename}")



import os

import cv2
import numpy as np
from numpy.typing import NDArray

from color_correction.constant.color_checker import reference_color_d50_bgr
from color_correction.constant.methods import (
    LiteralModelCorrection,
    LiteralModelDetection,
)
from color_correction.core.card_detection.det_yv8_onnx import (
    YOLOv8CardDetector,
)
from color_correction.core.correction import CorrectionModelFactory
from color_correction.processor.det_yv8 import DetectionProcessor
from color_correction.schemas.images import ColorPatchType, ImageType
from color_correction.utils.image_patch import (
    create_patch_tiled_image,
    visualize_patch_comparison,
)
from color_correction.utils.image_processing import calc_color_diff
from color_correction.utils.visualization_utils import (
    create_image_grid_visualization,
)


class ColorCorrection:
    """Color correction handler using color card detection and correction models.

    This class handles the complete workflow of color correction, including:
    - Color card detection in images
    - Color patch extraction
    - Color correction model training
    - Image correction application

    Parameters
    ----------
    detection_model : {'yolov8'}
        The model to use for color card detection.
    detection_conf_th : float, optional
        Confidence threshold for card detection.
    correction_model : {'least_squares', 'polynomial', 'linear_reg', 'affine_reg'}
        The model to use for color correction.
    reference_image : NDArray[np.uint8] | None, optional
        Reference image containing color checker card.
        If None, uses standard D50 values.
    use_gpu : bool, default=True
        Whether to use GPU for card detection. False will use CPU.
    **kwargs : dict
        Additional parameters for the correction model.

    Attributes
    ----------
    reference_patches : List[ColorPatchType] | None
        Extracted color patches from reference image.
    reference_grid_image : ImageType | None
        Visualization of reference color patches in grid format.
    reference_debug_image : ImageType | None
        Debug visualization of reference image preprocessing.
    """

    def __init__(
        self,
        detection_model: LiteralModelDetection = "yolov8",
        detection_conf_th: float = 0.25,
        correction_model: LiteralModelCorrection = "least_squares",
        reference_image: ImageType | None = None,
        use_gpu: bool = True,
        **kwargs: dict,
    ) -> None:
        # Initialize reference image attributes
        self.reference_patches = None
        self.reference_grid_image = None
        self.reference_debug_image = None

        # Initialize input image attributes
        self.input_patches = None
        self.input_grid_image = None
        self.input_debug_image = None

        # Initialize correction output attributes
        self.corrected_patches = None
        self.corrected_grid_image = None

        # Initialize model attributes
        self.trained_model = None
        self.correction_model = CorrectionModelFactory.create(
            model_name=correction_model,
            **kwargs,
        )
        self.card_detector = self._create_detector(
            model_name=detection_model,
            conf_th=detection_conf_th,
            use_gpu=use_gpu,
        )

        # Set reference patches
        self.set_reference_patches(image=reference_image)

    def _create_detector(
        self,
        model_name: str,
        conf_th: float = 0.25,
        use_gpu: bool = False,
    ) -> YOLOv8CardDetector:
        """Create a card detector instance.

        Parameters
        ----------
        model_name : str
            Name of the detector model to create.
        conf_th : float, optional
            Confidence threshold for card detection. Default is 0.25.
        use_gpu : bool, optional
            Whether to use GPU for detection. Default is False.

        Returns
        -------
        YOLOv8CardDetector
            Initialized detector instance.

        Raises
        ------
        ValueError
            If the model name is not supported.
        """
        if model_name != "yolov8":
            raise ValueError(f"Unsupported detection model: {model_name}")
        return YOLOv8CardDetector(use_gpu=use_gpu, conf_th=conf_th)

    def _extract_color_patches(
        self,
        image: ImageType,
        debug: bool = False,
    ) -> tuple[list[ColorPatchType], ImageType, ImageType | None]:
        """Extract color patches from an image using card detection.

        Parameters
        ----------
        image : ImageType
            Input image in BGR format.
        debug : bool, optional
            Whether to generate debug visualizations.

        Returns
        -------
        tuple[list[ColorPatchType], ImageType, ImageType | None]
            - List of BGR mean values for each detected patch
            - Grid visualization of detected patches
            - Debug visualization (if debug=True)
        """
        prediction = self.card_detector.detect(image=image)
        ls_bgr_mean_patch, grid_patch_img, debug_detection_viz = (
            DetectionProcessor.extract_color_patches(
                input_image=image,
                prediction=prediction,
                draw_processed_image=debug,
            )
        )
        return ls_bgr_mean_patch, grid_patch_img, debug_detection_viz

    def _save_debug_output(
        self,
        input_image: ImageType,
        corrected_image: ImageType,
        output_directory: str,
    ) -> None:
        """Save debug visualizations to disk.

        Parameters
        ----------
        input_image : ImageType
            The input image.
        corrected_image : ImageType
            The color-corrected image.
        output_directory : str
            Directory to save debug outputs.
        """
        before_comparison = visualize_patch_comparison(
            ls_mean_in=self.input_patches,
            ls_mean_ref=self.reference_patches,
        )
        after_comparison = visualize_patch_comparison(
            ls_mean_in=self.corrected_patches,
            ls_mean_ref=self.reference_patches,
        )

        # Create output directories
        run_dir = self._create_debug_directory(output_directory)

        # Prepare debug image grid
        image_collection = [
            ("Input Image", input_image),
            ("Corrected Image", corrected_image),
            ("Debug Preprocess", self.input_debug_image),
            ("Reference vs Input", before_comparison),
            ("Reference vs Corrected", after_comparison),
            ("[Free Space]", None),
            ("Patch Input", self.input_grid_image),
            ("Patch Corrected", self.corrected_grid_image),
            ("Patch Reference", self.reference_grid_image),
        ]

        # Save debug grid
        save_path = os.path.join(run_dir, "debug.jpg")
        create_image_grid_visualization(
            images=image_collection,
            grid_size=((len(image_collection) // 3) + 1, 3),
            figsize=(15, ((len(image_collection) // 3) + 1) * 4),
            save_path=save_path,
        )
        print(f"Debug output saved to: {save_path}")

    def _create_debug_directory(self, base_dir: str) -> str:
        """Create and return a unique debug output directory.

        Parameters
        ----------
        base_dir : str
            Base directory for debug outputs.

        Returns
        -------
        str
            Path to the created directory.
        """
        os.makedirs(base_dir, exist_ok=True)
        run_number = len(os.listdir(base_dir)) + 1
        run_dir = os.path.join(base_dir, f"{run_number}-{self.model_name}")
        os.makedirs(run_dir, exist_ok=True)
        return run_dir

    @property
    def model_name(self) -> str:
        "Return the name of the correction model."
        return self.correction_model.__class__.__name__

    @property
    def ref_patches(self) -> np.ndarray:
        """Return grid image of reference color patches."""
        return (
            self.reference_patches,
            self.reference_grid_image,
            self.reference_debug_image,
        )

    def set_reference_patches(
        self,
        image: np.ndarray | None,
        debug: bool = False,
    ) -> None:
        if image is None:
            self.reference_patches = reference_color_d50_bgr
            self.reference_grid_image = create_patch_tiled_image(self.reference_patches)
        else:
            (
                self.reference_patches,
                self.reference_grid_image,
                self.reference_debug_image,
            ) = self._extract_color_patches(image=image, debug=debug)

    def set_input_patches(self, image: np.ndarray, debug: bool = False) -> None:
        self.input_patches = None
        self.input_grid_image = None
        self.input_debug_image = None

        (
            self.input_patches,
            self.input_grid_image,
            self.input_debug_image,
        ) = self._extract_color_patches(image=image, debug=debug)
        return self.input_patches, self.input_grid_image, self.input_debug_image

    def fit(self) -> tuple[NDArray, list[ColorPatchType], list[ColorPatchType]]:
        """Fit color correction model using input and reference images.

        Parameters
        ----------
        input_image : NDArray
            Image BGR to be corrected that contains color checker classic 24 patches.
        reference_image : NDArray, optional
            Image BGR to be reference that contains color checker classic 24 patches.

        Returns
        -------
        Tuple[NDArray, List[NDArray], List[NDArray]]
            Correction weights, input patches, and reference patches.
        """
        if self.reference_patches is None:
            raise RuntimeError("Reference patches must be set before fitting model")

        if self.input_patches is None:
            raise RuntimeError("Input patches must be set before fitting model")

        self.trained_model = self.correction_model.fit(
            x_patches=self.input_patches,
            y_patches=self.reference_patches,
        )

        # Compute corrected patches
        self.corrected_patches = self.correction_model.compute_correction(
            input_image=np.array(self.input_patches),
        )
        self.corrected_grid_image = create_patch_tiled_image(self.corrected_patches)

        return self.trained_model

    def predict(
        self,
        input_image: ImageType,
        debug: bool = False,
        debug_output_dir: str = "output-debug",
    ) -> ImageType:
        """Apply color correction to input image.

        Parameters
        ----------
        input_image : ImageType
            Image to be color corrected.
        debug : bool, optional
            Whether to save debug visualizations.
        debug_output_dir : str, optional
            Directory to save debug outputs.

        Returns
        -------
        ImageType
            Color corrected image.

        Raises
        ------
        RuntimeError
            If model has not been fitted.
        """
        if self.trained_model is None:
            raise RuntimeError("Model must be fitted before correction")

        corrected_image = self.correction_model.compute_correction(
            input_image=input_image.copy(),
        )

        if debug:
            self._save_debug_output(
                input_image=input_image,
                corrected_image=corrected_image,
                output_directory=debug_output_dir,
            )

        return corrected_image

    def calc_color_diff_patches(self) -> dict:
        initial_color_diff = calc_color_diff(
            image1=self.input_grid_image,
            image2=self.reference_grid_image,
        )

        corrected_color_diff = calc_color_diff(
            image1=self.corrected_grid_image,
            image2=self.reference_grid_image,
        )

        delta_color_diff = {
            "min": initial_color_diff["min"] - corrected_color_diff["min"],
            "max": initial_color_diff["max"] - corrected_color_diff["max"],
            "mean": initial_color_diff["mean"] - corrected_color_diff["mean"],
            "std": initial_color_diff["std"] - corrected_color_diff["std"],
        }

        info = {
            "initial": initial_color_diff,
            "corrected": corrected_color_diff,
            "delta": delta_color_diff,
        }

        return info


if __name__ == "__main__":
    # Step 1: Define the path to the input image
    image_path = "asset/images/cc-19.png"
    image_path = "asset/images/cc-1.jpg"

    # Step 2: Load the input image
    input_image = cv2.imread(image_path)

    # Step 3: Initialize the color correction model with specified parameters
    color_corrector = ColorCorrection(
        detection_model="yolov8",
        detection_conf_th=0.25,
        correction_model="polynomial",
        # correction_model="least_squares",
        # correction_model="affine_reg",
        # correction_model="linear_reg",
        degree=3,  # for polynomial correction model
        use_gpu=True,
    )

    # Step 4: Extract color patches from the input image
    # you can set reference patches from another image (image has color checker card)
    # or use the default D50
    # color_corrector.set_reference_patches(image=None, debug=True)
    color_corrector.set_input_patches(image=input_image, debug=True)
    color_corrector.fit()
    corrected_image = color_corrector.predict(
        input_image=input_image,
        debug=True,
        debug_output_dir="zzz",
    )

    eval_result = color_corrector.calc_color_diff_patches()
    print(eval_result)



import os

import numpy as np
import pandas as pd

from color_correction.constant.methods import (
    LiteralModelCorrection,
    LiteralModelDetection,
)
from color_correction.services.color_correction import ColorCorrection
from color_correction.utils.image_patch import (
    visualize_patch_comparison,
)
from color_correction.utils.image_processing import calc_color_diff
from color_correction.utils.report_generator import ReportGenerator


class ColorCorrectionAnalyzer:
    def __init__(
        self,
        list_correction_methods: list[tuple[LiteralModelCorrection, dict]],
        list_detection_methods: list[tuple[LiteralModelDetection, dict]],
        use_gpu: bool = True,
    ) -> None:
        self.list_correction_methods = list_correction_methods
        self.list_detection_methods = list_detection_methods
        self.use_gpu = use_gpu
        self.rg = ReportGenerator()

    def _run_single_exp(
        self,
        idx: int,
        input_image: np.ndarray,
        det_method: LiteralModelDetection,
        det_params: dict,
        cc_method: LiteralModelCorrection,
        cc_params: dict,
        reference_image: np.ndarray | None = None,
    ) -> dict:
        cc = ColorCorrection(
            correction_model=cc_method,
            detection_model=det_method,
            detection_conf_th=det_params.get("detection_conf_th", 0.25),
            use_gpu=self.use_gpu,
            **cc_params,
        )

        if reference_image is not None:
            cc.set_reference_image(reference_image)
        cc.set_input_patches(input_image, debug=True)
        cc.fit()
        corrected_image = cc.predict(input_image=input_image)
        eval_results = cc.calc_color_diff_patches()

        before_comparison = visualize_patch_comparison(
            ls_mean_in=cc.input_patches,
            ls_mean_ref=cc.reference_patches,
        )
        after_comparison = visualize_patch_comparison(
            ls_mean_in=cc.corrected_patches,
            ls_mean_ref=cc.reference_patches,
        )

        dE_image = calc_color_diff(  # noqa: N806
            image1=input_image,
            image2=corrected_image,
        )

        one_row = {
            "Index": idx,
            "Detection Method": det_method,
            "Detection Parameters": det_params,
            "Drawed Preprocessing Input": cc.input_debug_image,
            "Drawed Preprocessing Reference": cc.reference_debug_image,
            "Correction Method": cc_method,
            "Correction Parameters": cc_params,
            "Color Patches - Before": before_comparison,
            "Color Patches - After": after_comparison,
            "Input Image": input_image,
            "Corrected Image": corrected_image,
            "Patch ΔE (Before) - Min": eval_results["initial"]["min"],
            "Patch ΔE (Before) - Max": eval_results["initial"]["max"],
            "Patch ΔE (Before) - Mean": eval_results["initial"]["mean"],
            "Patch ΔE (Before) - Std": eval_results["initial"]["std"],
            "Patch ΔE (After) - Min": eval_results["corrected"]["min"],
            "Patch ΔE (After) - Max": eval_results["corrected"]["max"],
            "Patch ΔE (After) - Mean": eval_results["corrected"]["mean"],
            "Patch ΔE (After) - Std": eval_results["corrected"]["std"],
            "Image ΔE - Min": dE_image["min"],
            "Image ΔE - Max": dE_image["max"],
            "Image ΔE - Mean": dE_image["mean"],
            "Image ΔE - Std": dE_image["std"],
        }
        return one_row

    def run(
        self,
        input_image: np.ndarray,
        output_dir: str = "benchmark_debug",
        reference_image: np.ndarray | None = None,
    ) -> pd.DataFrame:
        """
        Fungsi ini menjalankan benchmark untuk model color correction.
        """
        ls_data = []
        idx = 1
        for det_method, det_params in self.list_detection_methods:
            for cc_method, cc_params in self.list_correction_methods:
                print(
                    f"Running benchmark for {cc_method} method with {cc_params}",
                )
                data = self._run_single_exp(
                    idx=idx,
                    input_image=input_image,
                    det_method=det_method,
                    det_params=det_params,
                    cc_method=cc_method,
                    cc_params=cc_params,
                    reference_image=reference_image,
                )
                idx += 1
                ls_data.append(data)
        df_results = pd.DataFrame(ls_data)

        # Generate HTML report path
        os.makedirs(output_dir, exist_ok=True)
        html_report_path = os.path.join(output_dir, "report.html")
        pickel_report_path = os.path.join(output_dir, "report.pkl")

        # Report Generator -----------------------------------------------------
        self.rg.generate_html_report(df=df_results, path_html=html_report_path)
        self.rg.save_dataframe(df=df_results, filepath=pickel_report_path)

        # Save CSV report, but without image data
        df_results.drop(
            columns=[
                "Drawed Preprocessing Input",
                "Drawed Preprocessing Reference",
                "Color Patches - Before",
                "Color Patches - After",
                "Corrected Image",
                "Input Image",
            ],
        ).to_csv(os.path.join(output_dir, "report_no_image.csv"), index=False)

        print("DataFrame shape:", df_results.shape)
        print("\nDataFrame columns:", df_results.columns.tolist())


if __name__ == "__main__":
    # Pastikan path image sesuai dengan lokasi image Anda
    input_image_path = "asset/images/cc-19.png"

    benchmark = ColorCorrectionAnalyzer(
        list_correction_methods=[
            ("least_squares", {}),
            ("linear_reg", {}),
            ("affine_reg", {}),
            ("polynomial", {"degree": 2}),
            ("polynomial", {"degree": 3}),
            ("polynomial", {"degree": 4}),
        ],
        list_detection_methods=[
            ("yolov8", {"detection_conf_th": 0.25}),
        ],
    )

    benchmark.run(
        input_image_path,
        reference_image=None,
        output_dir="benchmark_debug",
    )



import numpy as np
import pytest
from color_correction.utils.correction import (
    preprocessing_compute,
    postprocessing_compute,
)

@pytest.mark.parametrize("input_image, expected_shape", [
    # When shape is exactly (24, 3) and should be cast without reshaping
    (np.random.rand(24, 3) * 255, (24, 3)),
    # When shape is different; for example a 6x6 image with 3 channels should be reshaped to (36, 3)
    (np.random.rand(6, 6, 3) * 255, (36, 3)),
])
def test_preprocessing_compute(input_image, expected_shape):
    output = preprocessing_compute(input_image)
    assert output.dtype == np.float32
    assert output.shape == expected_shape

@pytest.mark.parametrize("original_shape, predict_image, expected_shape", [
    # Grid image patches (original_shape length 2); no reshape needed
    ((24, 3), np.array([[300, -10, 100]] * 24), (24, 3)),
    # Regular colored image: reshape required from flat list to (h, w, c)
    ((4, 4, 3), np.array([[300, -10, 100]] * 16), (4, 4, 3)),
])
def test_postprocessing_compute(original_shape, predict_image, expected_shape):
    output = postprocessing_compute(original_shape, predict_image)
    # Output must be of type uint8 after clipping
    assert output.dtype == np.uint8
    assert output.shape == expected_shape
    # Validate that all values are clipped between 0 and 255
    assert output.min() >= 0
    assert output.max() <= 255



import numpy as np
from pydantic import BaseModel

from color_correction.utils.yolo_utils import draw_detections

BoundingBox = tuple[int, int, int, int]


class DetectionResult(BaseModel):
    boxes: list[BoundingBox]
    scores: list[float]
    class_ids: list[int]

    def draw_detections(
        self,
        image: np.ndarray,
    ) -> np.ndarray:
        """Draw detection boxes on image."""
        return draw_detections(image, self.boxes, self.scores, self.class_ids)



import time

import numpy as np

from color_correction.core.correction.base import BaseComputeCorrection
from color_correction.utils.correction import (
    postprocessing_compute,
    preprocessing_compute,
)


class LeastSquaresRegression(BaseComputeCorrection):
    def __init__(self) -> None:
        self.model = None

    def fit(
        self,
        x_patches: np.ndarray,  # input patches
        y_patches: np.ndarray,  # reference patches
    ) -> np.ndarray:
        start_time = time.perf_counter()

        self.model = np.linalg.lstsq(
            a=x_patches,
            b=y_patches,
            rcond=None,
        )[0]  # get only matrix of coefficients

        exc_time = time.perf_counter() - start_time
        print(f"{self.__class__.__name__} Fit: {exc_time} seconds")
        return self.model

    def compute_correction(self, input_image: np.ndarray) -> np.ndarray:
        if self.model is None:
            raise ValueError("Model is not fitted yet. Please call fit() method first.")

        # Input adalah array (N,3) dari nilai warna patches
        org_input_shape = input_image.shape
        input_image = preprocessing_compute(input_image)
        image = np.dot(input_image, self.model)
        corrected_image = postprocessing_compute(org_input_shape, image)
        return corrected_image



import numpy as np


def create_patch_tiled_image(
    ls_patches: list[tuple[int, int, int, int]],
    patch_size: tuple[int, int, int] = (50, 50, 1),
) -> np.ndarray:
    """Generate a color patch image from a list of BGR values.

    This function creates a color patch image by tiling BGR values into patches and
    arranging them in a 4x6 grid pattern. Each patch is repeated according to the
    specified patch size.

    Parameters
    ----------
    ls_patches : list of tuple
        List containing 24 BGR color tuples, where each tuple has three integers
        representing (B, G, R) values.
    patch_size : tuple of int, optional
        Size of each individual patch in pixels, by default (50, 50, 1).
        Format is (height, width, channels).

    Returns
    -------
    numpy.ndarray
        Generated image as a numpy array with shape determined by patch_size and
        arrangement (4 rows x 6 columns). Array type is uint8.

    Notes
    -----
    This function is specifically designed to work with 24 color patches arranged
    in a 4x6 grid pattern.

    Examples
    --------
    >>> patches = [(255, 0, 0), (0, 255, 0), ...] # 24 BGR tuples
    >>> patch_size = (50, 50, 1)
    >>> image = generate_image_patches(patches, patch_size)
    """

    ls_stack_h = []
    ls_stack_v = []

    if len(ls_patches) != 24:
        raise ValueError("Failed to generate image. The number of patches must be 24.")

    for _idx, patch in enumerate(ls_patches, start=1):
        patch_img = np.tile(patch, patch_size)
        ls_stack_h.append(patch_img)
        if _idx % 6 == 0:
            row = np.hstack(ls_stack_h)
            ls_stack_v.append(row)
            ls_stack_h = []
    image = np.vstack(ls_stack_v).astype(np.uint8)
    return image


def visualize_patch_comparison(
    ls_mean_ref: np.ndarray,
    ls_mean_in: np.ndarray,
    patch_size: tuple[int, int, int] = (100, 100, 1),
) -> np.ndarray:
    """
    Compare two sets of image patches by inserting a resized inner patch into
    the center of an outer patch. This visualization grid helps in comparing
    the reference and input images in a structured manner.

    Parameters
    ----------
    ls_mean_ref : list of np.ndarray
        List of outer image patches. Each patch is repeated to form the full
        grid background.
    ls_mean_in : list of np.ndarray
        List of inner image patches meant to be resized and placed into the
        center of the outer patches.
    patch_size : tuple of int, optional
        A tuple specifying the size of the patch in the format (height, width, channels)
        by default (100, 100, 1).

    Returns
    -------
    np.ndarray
        The final composited image with each outer patch modified with the
        corresponding resized inner patch, arranged in a grid format.
    """

    ls_stack_h = []
    ls_stack_v = []

    h = patch_size[0]
    w = patch_size[1]
    h_2 = h // 2
    w_2 = w // 2
    y1 = h_2 - (h // 4) - 1
    y2 = h_2 + (h // 4)
    x1 = w_2 - (w // 4) - 1
    x2 = w_2 + (w // 4)

    for _idx, (patch_ref, patch_in) in enumerate(
        zip(ls_mean_ref, ls_mean_in, strict=False),
        start=1,
    ):
        img_patch_ref = np.tile(patch_ref, patch_size)
        img_patch_in = np.tile(
            patch_in,
            (y2 - y1, x2 - x1, patch_size[2]),
        )

        # img_patch_in = cv2.resize(img_patch_in, (y2 - y1, x2 - x1))
        img_patch_ref[y1:y2, x1:x2] = img_patch_in
        ls_stack_h.append(img_patch_ref)

        if _idx % 6 == 0:
            row = np.hstack(ls_stack_h)
            ls_stack_v.append(row)
            ls_stack_h = []
    image = np.vstack(ls_stack_v).astype(np.uint8)
    return image



import platform
import subprocess
from typing import Any

from color_correction.schemas.device import (
    CPUArchitecture,
    DeviceSpecs,
    GPUType,
)


def detect_darwin(specs: dict[str, Any]) -> dict[str, Any]:
    """Detect hardware specifications for macOS (Darwin).

    Parameters
    ----------
    specs : dict
        Base specifications dictionary with OS information.

    Returns
    -------
    dict
        Updated specifications with CPU and GPU information.

    Notes
    -----
    Detects:
    1. Apple Silicon vs Intel via sysctl
    2. GPU type (Apple/AMD/NVIDIA) via system_profiler
    """
    try:
        cpu_info = subprocess.check_output(
            ["sysctl", "-n", "machdep.cpu.brand_string"],
        ).decode()
        specs["is_apple_silicon"] = "Apple" in cpu_info
        specs["cpu_arch"] = (
            CPUArchitecture.APPLE if "Apple" in cpu_info else CPUArchitecture.INTEL
        )
    except subprocess.SubprocessError:
        specs["cpu_arch"] = CPUArchitecture.UNKNOWN

    try:
        gpu_info = subprocess.check_output(
            ["system_profiler", "SPDisplaysDataType"],
        ).decode()
        if "Apple M" in gpu_info:
            specs["gpu_type"] = GPUType.APPLE
        elif "AMD" in gpu_info:
            specs["gpu_type"] = GPUType.AMD
        elif "NVIDIA" in gpu_info:
            specs["gpu_type"] = GPUType.NVIDIA
        else:
            specs["gpu_type"] = GPUType.UNKNOWN
    except subprocess.SubprocessError:
        specs["gpu_type"] = GPUType.UNKNOWN

    return specs


def detect_linux(specs: dict[str, Any]) -> dict[str, Any]:
    """Detect hardware specifications for Linux.

    Parameters
    ----------
    specs : dict
        Base specifications dictionary with OS information.

    Returns
    -------
    dict
        Updated specifications with CPU and GPU information.

    Notes
    -----
    Detects:
    1. CPU architecture (ARM/x86_64) via lscpu
    2. GPU type (NVIDIA/AMD) via nvidia-smi or lspci
    """
    try:
        cpu_info = subprocess.check_output("lscpu", shell=True).decode().lower()
        specs["cpu_arch"] = (
            CPUArchitecture.ARM if "aarch64" in cpu_info else CPUArchitecture.INTEL
        )
    except subprocess.SubprocessError:
        specs["cpu_arch"] = CPUArchitecture(platform.machine())

    # GPU detection
    try:
        subprocess.check_output("nvidia-smi", shell=True)
        specs["gpu_type"] = GPUType.NVIDIA
    except subprocess.SubprocessError:
        try:
            amd_info = (
                subprocess.check_output(
                    "lspci | grep -i amd",
                    shell=True,
                )
                .decode()
                .lower()
            )
            specs["gpu_type"] = (
                GPUType.AMD
                if ("vga" in amd_info or "amd" in amd_info)
                else GPUType.UNKNOWN
            )
        except subprocess.SubprocessError:
            specs["gpu_type"] = GPUType.UNKNOWN

    return specs


def detect_windows(specs: dict[str, Any]) -> dict[str, Any]:
    """Detect hardware specifications for Windows.

    Parameters
    ----------
    specs : dict
        Base specifications dictionary with OS information.

    Returns
    -------
    dict
        Updated specifications with CPU and GPU information.
    """
    proc = platform.processor().lower()
    if "intel" in proc:
        specs["cpu_arch"] = CPUArchitecture.INTEL
    elif "amd" in proc:
        specs["cpu_arch"] = CPUArchitecture.AMD
    elif "arm" in proc:
        specs["cpu_arch"] = CPUArchitecture.ARM
    else:
        specs["cpu_arch"] = CPUArchitecture.UNKNOWN

    specs["gpu_type"] = GPUType.UNKNOWN
    return specs


def get_device_specs() -> DeviceSpecs:
    """Get device hardware specifications.

    Returns
    -------
    DeviceSpecs
        Pydantic model containing device specifications.
    """
    specs = {
        "os_name": platform.system(),
        "cpu_arch": CPUArchitecture.UNKNOWN,
        "gpu_type": GPUType.UNKNOWN,
        "is_apple_silicon": False,
    }

    detector_map = {
        "Darwin": detect_darwin,
        "Linux": detect_linux,
        "Windows": detect_windows,
    }

    detect_func = detector_map.get(specs["os_name"], lambda x: x)
    specs = detect_func(specs)

    return DeviceSpecs(**specs)


if __name__ == "__main__":
    specs = get_device_specs()
    print(f"Detected specs: {specs.model_dump()}")



from enum import Enum

from pydantic import BaseModel, Field


class GPUType(str, Enum):
    NVIDIA = "NVIDIA"
    AMD = "AMD"
    APPLE = "Apple Integrated"
    UNKNOWN = "Unknown GPU"


class CPUArchitecture(str, Enum):
    INTEL = "Intel"
    AMD = "AMD"
    ARM = "ARM"
    APPLE = "Apple Silicon"
    UNKNOWN = "Unknown"


class DeviceSpecs(BaseModel):
    """Device specifications schema."""

    os_name: str = Field(..., description="Operating system name")
    cpu_arch: CPUArchitecture = Field(
        CPUArchitecture.UNKNOWN,
        description="CPU architecture",
    )
    gpu_type: GPUType = Field(GPUType.UNKNOWN, description="GPU type")
    is_apple_silicon: bool = Field(False, description="Whether device is Apple Silicon")



import numpy as np
import pytest

from color_correction.utils.image_patch import (
    create_patch_tiled_image,
    visualize_patch_comparison,
)


@pytest.fixture
def sample_patches_bgr() -> list[tuple[int, int, int]]:
    return np.random.randint(0, 255, size=(24, 3))


@pytest.fixture
def sample_outer_patches() -> list[np.ndarray]:
    return np.random.randint(0, 255, size=(24, 3))


@pytest.fixture
def sample_inner_patches() -> list[np.ndarray]:
    return np.random.randint(0, 255, size=(24, 3))


def test_create_patch_tiled_image(sample_patches_bgr) -> None:  # noqa: ANN001
    patch_size = (50, 50, 1)
    image = create_patch_tiled_image(sample_patches_bgr, patch_size)
    assert image.shape == (4 * patch_size[0], 6 * patch_size[1], 3 * patch_size[2])
    assert image.dtype == np.uint8


def test_create_patch_tiled_image_custom_size(sample_patches_bgr) -> None:  # noqa: ANN001
    patch_size = (30, 30, 1)
    image = create_patch_tiled_image(sample_patches_bgr, patch_size)
    assert image.shape == (4 * patch_size[0], 6 * patch_size[1], 3 * patch_size[2])


def test_visualize_patch_comparison(sample_outer_patches, sample_inner_patches):
    patch_size = (100, 100, 1)
    image = visualize_patch_comparison(
        sample_outer_patches,
        sample_inner_patches,
        patch_size,
    )
    assert image.shape == (4 * patch_size[0], 6 * patch_size[1], 3 * patch_size[2])
    assert image.dtype == np.uint8


def test_visualize_inner_patch_center(sample_outer_patches, sample_inner_patches):
    patch_size = (100, 100, 1)
    h, w, _ = patch_size
    h_half = h // 2
    w_half = w // 2
    y1 = h_half - (h // 4)
    y2 = h_half + (h // 4)
    x1 = w_half - (w // 4)
    x2 = w_half + (w // 4)

    image = visualize_patch_comparison(
        sample_outer_patches,
        sample_inner_patches,
        patch_size,
    )
    inner_patch = sample_inner_patches[0]
    inner_patch = np.tile(inner_patch, (y2 - y1, x2 - x1, 1))
    assert np.array_equal(image[y1:y2, x1:x2], inner_patch)



import cv2
import matplotlib.figure
import matplotlib.pyplot as plt
import numpy as np


def create_image_grid_visualization(
    images: list[tuple[str, np.ndarray | matplotlib.figure.Figure | None]],
    grid_size: tuple[int, int] = (2, 3),
    figsize: tuple[int, int] = (15, 10),
    save_path: str | None = None,
    dpi: int = 300,
) -> matplotlib.figure.Figure:
    """
    Display images in a grid layout with titles

    Parameters:
    -----------
    images : List[Tuple[str, Union[np.ndarray, matplotlib.figure.Figure, None]]]
        List of tuples containing (title, image)
    grid_size : Tuple[int, int]
        Grid layout in (rows, columns) format
    figsize : Tuple[int, int]
        Size of the entire figure in inches
    save_path : Optional[str]
        If provided, save the figure to this path
    dpi : int
        DPI for saved figure

    Returns:
    --------
    matplotlib.figure.Figure
        The figure object containing the grid
    """

    rows, cols = grid_size
    fig = plt.figure(figsize=figsize)

    for idx, (title, img) in enumerate(images):
        if idx >= rows * cols:
            print(
                f"Warning: Only showing first {rows * cols} images due to "
                "grid size limitation",
            )
            break

        ax = fig.add_subplot(rows, cols, idx + 1)

        # Handle different image types
        if isinstance(img, np.ndarray):
            if len(img.shape) == 2:  # Grayscale
                ax.imshow(img, cmap="gray")
            else:  # RGB/RGBA
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                ax.imshow(img)
        elif isinstance(img, matplotlib.figure.Figure):
            # Convert matplotlib figure to image array
            fig.canvas.draw()
            img_array = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
            img_array = img_array.reshape(fig.canvas.get_width_height()[::-1] + (3,))
            ax.imshow(img_array)

        ax.set_title(title)
        ax.axis("off")

    plt.tight_layout()

    # Save figure if path is provided
    if save_path:
        fig.savefig(save_path, dpi=dpi, bbox_inches="tight")
        print(f"Figure saved to: {save_path}")

    plt.close()  # Close the figure to free memory
    return fig



from color_correction.core.correction.affine_reg import AffineReg
from color_correction.core.correction.least_squares import (
    LeastSquaresRegression,
)
from color_correction.core.correction.linear_reg import LinearReg
from color_correction.core.correction.polynomial import Polynomial


class CorrectionModelFactory:
    @staticmethod
    def create(model_name: str, **kwargs: dict) -> ...:
        model_registry = {
            "least_squares": LeastSquaresRegression(),
            "polynomial": Polynomial(**kwargs),
            "linear_reg": LinearReg(),
            "affine_reg": AffineReg(),
        }
        return model_registry.get(model_name)



import numpy as np
import pytest

from color_correction.core.card_detection.det_yv8_onnx import YOLOv8CardDetector

@pytest.mark.skip(reason="Test is not implemented")
def test_detector_init(sample_image: np.ndarray) -> None:
    detector = YOLOv8CardDetector(use_gpu=False)
    result = detector.detect(sample_image)
    assert result is not None
    assert len(result.boxes) == 0  # Expect no detections on empty image



import cv2
import numpy as np

from color_correction.schemas.det_yv8 import DetectionResult
from color_correction.utils.geometry_processing import (
    extract_intersecting_patches,
    generate_expected_patches,
    suggest_missing_patch_coordinates,
)
from color_correction.utils.image_processing import (
    calc_mean_color_patch,
    crop_region_with_margin,
)

# Type aliases for better readability
BoundingBox = tuple[int, int, int, int]
RGB = tuple[float, float, float]
BGR = tuple[float, float, float]


class DetectionProcessor:
    """
    A class to process color calibration card detections and extract color patches.

    This class handles the detection and processing of color calibration cards and their
    individual color patches, including visualization and RGB value extraction.
    """

    @staticmethod
    def get_each_class_box(
        prediction: DetectionResult,
    ) -> tuple[list[BoundingBox], list[BoundingBox]]:
        """
        Separate detection boxes by class (cards and patches).

        Parameters
        ----------
        prediction : DetectionResult
            Detection results containing boxes and class IDs

        Returns
        -------
        Tuple[List[BoundingBox], List[BoundingBox]]
            Two lists containing card boxes and patch boxes respectively
        """
        ls_cards = [
            box
            for box, class_id in zip(
                prediction.boxes,
                prediction.class_ids,
                strict=False,
            )
            if class_id == 1
        ]
        ls_patches = [
            box
            for box, class_id in zip(
                prediction.boxes,
                prediction.class_ids,
                strict=False,
            )
            if class_id == 0
        ]
        return ls_cards, ls_patches

    @staticmethod
    def print_summary(prediction: DetectionResult) -> None:
        """
        Print a summary of detected objects.

        Parameters
        ----------
        prediction : DetectionResult
            Detection results to summarize
        """
        ls_cards, ls_patches = DetectionProcessor.get_each_class_box(prediction)
        print(f"Number of cards detected: {len(ls_cards)}")
        print(f"Number of patches detected: {len(ls_patches)}")

    @staticmethod
    def process_patches(
        input_image: np.ndarray,
        ordered_patches: list[tuple[BoundingBox, tuple[int, int]] | None],
    ) -> tuple[list[RGB], np.ndarray]:
        """
        Process detected patches to extract RGB values and create a visualization.

        Parameters
        ----------
        input_image : np.ndarray
            Input image containing the patches
        ordered_patches : List[Optional[BoundingBox]]
            List of ordered patch coordinates

        Returns
        -------
        Tuple[List[RGB], np.ndarray]
            List of RGB values and visualization image
        """
        patch_size = (50, 50, 1)
        ls_bgr_mean_patch = []
        ls_horizontal_patch = []
        ls_vertical_patch = []

        for idx, coord_patch in enumerate(ordered_patches, start=1):
            if coord_patch is None:
                continue

            bbox_patch, _ = coord_patch

            # Extract and process each patch
            cropped_patch = crop_region_with_margin(
                image=input_image,
                coordinates=bbox_patch,
                margin_ratio=0.2,
            )
            bgr_mean_patch = calc_mean_color_patch(cropped_patch)
            ls_bgr_mean_patch.append(bgr_mean_patch)

            # Build visualization
            patch_viz = np.tile(bgr_mean_patch, patch_size)
            ls_horizontal_patch.append(patch_viz)
            if idx % 6 == 0:
                ls_vertical_patch.append(np.hstack(ls_horizontal_patch))
                ls_horizontal_patch = []

        patches_image = np.vstack(ls_vertical_patch)
        return ls_bgr_mean_patch, patches_image

    @staticmethod
    def extract_color_patches(
        input_image: np.ndarray,
        prediction: DetectionResult,
        draw_processed_image: bool = False,
    ) -> tuple[list[BGR], np.ndarray, np.ndarray | None]:
        """
        Extract and process color patches from the detected calibration card.

        Parameters
        ----------
        input_image : np.ndarray
            Input image containing the color calibration card
        prediction : DetectionResult
            Detection results from YOLOv8
        draw_processed_image : bool, optional
            Whether to create a visualization image, by default False

        Returns
        -------
        Tuple[List[BGR], np.ndarray, Optional[np.ndarray]]
            BGR values, patch visualization, and optional detection visualization

        Raises
        ------
        ValueError
            If no cards or patches are detected
        """
        ls_cards, ls_patches = DetectionProcessor.get_each_class_box(prediction)

        if not ls_cards:
            raise ValueError("No cards detected")
        if not ls_patches:
            raise ValueError("No patches detected")

        # Generate expected patch grid
        card_box = ls_cards[0]
        ls_grid_card = generate_expected_patches(card_box)

        # Match detected patches with grid
        ls_ordered_patch_bbox = extract_intersecting_patches(
            ls_patches=ls_patches,
            ls_grid_card=ls_grid_card,
        )

        # Handle missing patches
        d_suggest = None
        if None in ls_ordered_patch_bbox:
            print("Auto filling missing patches...")
            ls_ordered_bbox_only = [
                patch[0] if patch is not None else None
                for patch in ls_ordered_patch_bbox
            ]
            d_suggest = suggest_missing_patch_coordinates(ls_ordered_bbox_only)
            for idx, patch in d_suggest.items():
                cxpatch = (patch[0] + patch[2]) // 2
                cypatch = (patch[1] + patch[3]) // 2
                ls_ordered_patch_bbox[idx] = (patch, (cxpatch, cypatch))

        # Process patches and create visualizations
        ls_bgr_mean_patch, grid_patch_img = DetectionProcessor.process_patches(
            input_image=input_image,
            ordered_patches=ls_ordered_patch_bbox,
        )

        detection_viz = None
        if draw_processed_image:
            detection_viz = DetectionProcessor.draw_preprocess(
                image=input_image,
                expected_boxes=ls_grid_card,
                prediction=prediction,
                ls_ordered_patch_bbox=ls_ordered_patch_bbox,
                suggested_patches=d_suggest,
            )

        return ls_bgr_mean_patch, grid_patch_img, detection_viz

    @staticmethod
    def draw_preprocess(
        image: np.ndarray,
        expected_boxes: list[BoundingBox],
        prediction: DetectionResult,
        ls_ordered_patch_bbox: list[BoundingBox | None],
        suggested_patches: dict[int, BoundingBox] | None = None,
    ) -> np.ndarray:
        """
        Draw detection visualizations on the image.

        Parameters
        ----------
        image : np.ndarray
            Input image to draw on
        boxes : List[BoundingBox]
            List of bounding boxes to draw
        patch_indices : Optional[List[int]]
            Indices to label the patches
        suggested_patches : Optional[Dict[int, BoundingBox]]
            Additional suggested patch locations to draw

        Returns
        -------
        np.ndarray
            Image with visualizations
        """
        color_green = (0, 255, 0)
        color_cyan = (255, 255, 10)
        color_violet = (255, 0, 255)
        color_red = (0, 0, 255)
        color_blue = (255, 0, 0)

        result_image = image.copy()

        # Draw all expected boxes
        for idx_b, box in enumerate(expected_boxes):
            cv2.rectangle(
                img=result_image,
                pt1=(box[0], box[1]),
                pt2=(box[2], box[3]),
                color=color_green,
                thickness=2,
            )

            # draw connection lines between expected and intersecting patches
            patch = ls_ordered_patch_bbox[idx_b]
            if patch is None:
                continue
            cx, cy = patch[1]
            crefx, crefy = (box[0] + box[2]) // 2, (box[1] + box[3]) // 2
            cv2.line(
                img=result_image,
                pt1=(cx, cy),
                pt2=(crefx, crefy),
                color=color_blue,
                thickness=1,
            )

        # draw all predicted boxes
        for pbox, pids, pscore in zip(
            prediction.boxes,
            prediction.class_ids,
            prediction.scores,
            strict=False,
        ):
            if pids == 1:
                continue
            cv2.rectangle(
                img=result_image,
                pt1=(pbox[0], pbox[1]),
                pt2=(pbox[2], pbox[3]),
                color=color_cyan,
                thickness=2,
            )
            cv2.putText(
                img=result_image,
                text=f"{pids} {pscore:.2f}",
                org=(pbox[0] + 3, pbox[1] + 12),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=0.4,
                color=color_red,
                thickness=1,
                lineType=cv2.LINE_AA,
            )

        # Draw suggested patches if provided
        if suggested_patches:
            for box in suggested_patches.values():
                cv2.rectangle(
                    img=result_image,
                    pt1=(box[0], box[1]),
                    pt2=(box[2], box[3]),
                    color=color_violet,
                    thickness=2,
                )

        return result_image



# report_generator.py
from datetime import datetime
from importlib import resources

import pandas as pd

from color_correction.utils.formater import format_value


class ReportGenerator:
    def __init__(self) -> None:
        self.template_dir = "color_correction.templates"

    def _read_template(self, filename: str) -> str:
        """Read template file content"""
        with resources.files(self.template_dir).joinpath(filename).open("r") as f:
            return f.read()

    def generate_report(self, body_report: str) -> str:
        """Generate full HTML report"""
        # Load components
        styles = self._read_template("style-report.css")
        scripts = self._read_template("script-report.js")
        column_desc = self._read_template("column_description.html")
        base_template = self._read_template("base_report.html")

        current_time = datetime.now().strftime("%d %B %Y, %H:%M:%S")

        # Replace placeholders
        final_html = base_template.format(
            styles=styles,
            scripts=scripts,
            current_time=current_time,
            body_report=body_report,
            column_descriptions=column_desc,
        )
        return final_html

    def generate_table(self, headers: list, rows: list) -> str:
        """Generate table HTML"""
        headers_html = "".join([f"<th>{h}</th>" for h in headers])
        rows_html = "".join([f"<tr>{r}</tr>" for r in rows])

        return f"""
        <table>
            <thead>
                <tr>{headers_html}</tr>
            </thead>
            <tbody>
                {rows_html}
            </tbody>
        </table>
        """

    def generate_html_report(self, df: pd.DataFrame, path_html: str) -> str:
        """Generate HTML report from DataFrame with images"""
        df_html = df.copy()

        # Generate rows
        rows = []
        for _, row in df_html.iterrows():
            row_html = "<tr>"
            for value in row:
                formatted_value = format_value(value)
                row_html += f"<td>{formatted_value}</td>"
            row_html += "</tr>"
            rows.append(row_html)

        table_html = self.generate_table(
            headers=df.columns.tolist(),
            rows=rows,
        )
        report_html = self.generate_report(table_html)

        with open(path_html, "w") as f:
            f.write(report_html)

        return report_html

    def save_dataframe(self, df: pd.DataFrame, filepath: str) -> None:
        """Save DataFrame with pickle to preserve numpy arrays"""
        df.to_pickle(filepath)



import numpy as np


def preprocessing_compute(input_image: np.ndarray) -> np.ndarray:
    if input_image.shape == (24, 3):
        # to handle grid image patches only
        image = input_image.astype(np.float32)
    else:
        image = input_image.reshape(-1, 3).astype(np.float32)
    return image


def postprocessing_compute(
    original_shape: tuple,
    predict_image: np.ndarray,
) -> np.ndarray:
    if len(original_shape) == 2:
        # to handle grid image patches only
        corrected_image = np.clip(predict_image, 0, 255).astype(np.uint8)
    else:
        h, w, c = original_shape
        corrected_image = (
            np.clip(predict_image, 0, 255).astype(np.uint8).reshape(h, w, c)
        )
    return corrected_image



import numpy as np
import pytest
from color_correction.utils.image_processing import crop_region_with_margin, calc_mean_color_patch

@pytest.fixture
def known_image() -> np.ndarray:
    # Create an image with a known pattern using np.arange,
    # reshape to (100, 100, 3) and wrap values with modulo 255
    img = np.arange(100 * 100 * 3, dtype=np.uint8) % 255
    return img.reshape((100, 100, 3))

@pytest.mark.parametrize(
    "coordinates, margin_ratio, expected_slice",
    [
        (
            (10, 20, 90, 80),
            0.2,
            (slice(32, 68), slice(26, 74))
        ),
        (
            (10, 20, 90, 80),
            0.0,
            (slice(20, 80), slice(10, 90))
        ),
    ],
)
def test_crop_region_with_margin(known_image: np.ndarray, coordinates, margin_ratio, expected_slice) -> None:
    result = crop_region_with_margin(known_image, coordinates, margin_ratio)
    expected = known_image[expected_slice[0], expected_slice[1]]
    np.testing.assert_array_equal(result, expected)

@pytest.mark.parametrize(
    "img, expected_mean",
    [
        (
            np.full((10, 10, 3), fill_value=50, dtype=np.uint8),
            np.array([50, 50, 50], dtype=np.uint8)
        ),
        (
            # Create an image where channel 0 is 0, channel 1 is 100, channel 2 is 200.
            np.stack([
                np.zeros((20, 20), dtype=np.uint8),
                np.full((20, 20), 100, dtype=np.uint8),
                np.full((20, 20), 200, dtype=np.uint8)
            ], axis=-1),
            np.array([0, 100, 200], dtype=np.uint8)
        ),
    ],
)
def test_calc_mean_color_patch(img: np.ndarray, expected_mean) -> None:
    mean_color = calc_mean_color_patch(img)
    np.testing.assert_array_equal(mean_color, expected_mean)



import base64
import io

import colour as cl
import cv2
import numpy as np
from numpy.typing import NDArray
from PIL import Image

ImageType = NDArray[np.uint8]


def crop_region_with_margin(
    image: np.ndarray,
    coordinates: tuple[int, int, int, int],
    margin_ratio: float = 0.2,
) -> np.ndarray:
    """Crop a region from image with additional margin from given coordinates.

    Parameters
    ----------
    image : np.ndarray
        Input image array of shape (H, W, C) or (H, W).
    coordinates : np.ndarray
        Bounding box coordinates [x1, y1, x2, y2].
    margin_ratio : float, optional
        Ratio of margin to add relative to region size, by default 0.2.

    Returns
    -------
    np.ndarray
        Cropped image region with margins.
    """
    y1, y2 = coordinates[1], coordinates[3]
    x1, x2 = coordinates[0], coordinates[2]

    height = y2 - y1
    margin_y = height * margin_ratio
    width = x2 - x1
    margin_x = width * margin_ratio

    crop_y1 = int(y1 + margin_y)
    crop_y2 = int(y2 - margin_y)
    crop_x1 = int(x1 + margin_x)
    crop_x2 = int(x2 - margin_x)

    return image[crop_y1:crop_y2, crop_x1:crop_x2]


def calc_mean_color_patch(img: np.ndarray) -> np.ndarray:
    """Calculate mean RGB/BGR values across spatial dimensions.

    Parameters
    ----------
    img : np.ndarray
        Input image array of shape (H, W, C).

    Returns
    -------
    np.ndarray
        Array of mean RGB values, shape (C,), dtype uint8.
    """
    return np.mean(img, axis=(0, 1)).astype(np.uint8)


def calc_color_diff(
    image1: ImageType,
    image2: ImageType,
) -> dict[str, float]:
    """Calculate color difference metrics between two images.

    Parameters
    ----------
    image1, image2 : NDArray
        Images to compare in BGR format.

    Returns
    -------
    dict[str, float]
        Dictionary of color difference
        keys: min, max, mean, std
    """
    rgb1 = cv2.cvtColor(image1, cv2.COLOR_BGR2RGB)
    rgb2 = cv2.cvtColor(image2, cv2.COLOR_BGR2RGB)

    lab1 = cl.XYZ_to_Lab(cl.sRGB_to_XYZ(rgb1 / 255))
    lab2 = cl.XYZ_to_Lab(cl.sRGB_to_XYZ(rgb2 / 255))

    delta_e = cl.difference.delta_E(lab1, lab2, method="CIE 2000")

    return {
        "min": round(float(np.min(delta_e)), 4),
        "max": round(float(np.max(delta_e)), 4),
        "mean": round(float(np.mean(delta_e)), 4),
        "std": round(float(np.std(delta_e)), 4),
    }


def numpy_array_to_base64(
    arr: np.ndarray,
    convert_bgr_to_rgb: bool = True,
) -> str:
    """Convert numpy array (image) to base64 string"""
    if arr is None:
        return ""

    if convert_bgr_to_rgb:
        arr = cv2.cvtColor(arr, cv2.COLOR_BGR2RGB)
    img = Image.fromarray(arr)
    buffer = io.BytesIO()
    img.save(buffer, format="PNG")
    img_str = base64.b64encode(buffer.getvalue()).decode()
    return f"data:image/png;base64,{img_str}"
